import requests
from tornado.escape import json_encode, json_decode
from tornado.httpclient import HTTPRequest
from tornado.ioloop import IOLoop
from tornado.websocket import websocket_connect
from uuid import uuid4

DEFAULT_TIMEOUT = 60 * 60
DEFAULT_USERNAME = 'anonymous'


class KernelLauncher:

    def __init__(self, host):
        self.http_api_endpoint = 'http://{}/api/kernels'.format(host)
        self.ws_api_endpoint = 'ws://{}/api/kernels'.format(host)

    def launch(self, kernelspec_name, username=DEFAULT_USERNAME):
        print('Launching up a {} kernel on the Gateway....'.format(kernelspec_name))
        kernel_id = None
        json_data = {'name': kernelspec_name}
        if username is not None:
            json_data['env'] = {'KERNEL_USERNAME': username}
        response = requests.post(self.http_api_endpoint, data=json_encode(json_data))
        if response.status_code == 201:
            json_data = response.json()
            kernel_id = json_data.get("id")
            print('Launched kernel {}'.format(kernel_id))
        else:
            raise RuntimeError('Error creating kernel : {} response code \n {}'.format(response.status_code, response.content))

        return Kernel(self.ws_api_endpoint, kernel_id)

    def shutdown(self, kernel_id):
        print("Shutting down kernel : {}".format(kernel_id))
        if not kernel_id:
            return False
        url = "{}/{}".format(self.http_api_endpoint, kernel_id)
        response = requests.delete(url)
        if response.status_code == 204:
            print('Kernel {} shutdown'.format(kernel_id))
            return True
        else:
            raise RuntimeError('Error shutting down kernel {}: {}'.format(kernel_id, response.content))


class Kernel:

    def __init__(self, ws_api_endpoint, kernel_id):
        self.ws_api_endpoint = ws_api_endpoint
        self.kernel_api_endpoint = '{}/{}/channels'.format(ws_api_endpoint, kernel_id)
        self.kernel_id = kernel_id
        print('Initializing kernel client ({}) to {}'.format(kernel_id, self.kernel_api_endpoint))


    def __create_execute_request(self, msg_id, code):
        return json_encode({
            'header': {
                'username': '',
                'version': '5.0',
                'session': '',
                'msg_id': msg_id,
                'msg_type': 'execute_request'
            },
            'parent_header': {},
            'channel': 'shell',
            'content': {
                'code': code,
                'silent': False,
                'store_history': False,
                'user_expressions': {},
                'allow_stdin': False
            },
            'metadata': {},
            'buffers': {}
        })

    def execute(self, code, timeout=DEFAULT_TIMEOUT):

        kernel_socket = None
        response_type = None
        response = []

        #print('')
        #print('Submitting code:\n{}'.format(code))
        #print('')

        try:
            ws_req = HTTPRequest(self.kernel_api_endpoint)
            kernel_socket_future = websocket_connect(ws_req)
            kernel_socket = IOLoop.current().run_sync(lambda: kernel_socket_future, timeout)

            #print('Sending message to kernel')
            msg_id = uuid4().hex
            message = self.__create_execute_request(msg_id, code)
            #pprint(message)
            future = kernel_socket.write_message(message)
            response_message = IOLoop.current().run_sync(lambda: future, timeout)

            while True:
                msg_future = kernel_socket.read_message()
                response_message = IOLoop.current().run_sync(lambda: msg_future, timeout)

                response_message = json_decode(response_message)

                #print('Received message from kernel')
                #pprint(response_message)

                response_message_id = response_message['parent_header']['msg_id']
                response_message_type = response_message['msg_type']

                if response_message_type == 'error':
                    raise RuntimeError('ERROR: {}:{}'.format(response_message['content']['ename'], response_message['content']['evalue']))

                if response_message_type == 'stream':
                    response_type = 'text'
                    response.append(response_message['content']['text'])

                if response_message_type == 'execute_result':
                    if 'text/plain' in response_message['content']['data']:
                        response_type = 'text'
                        response.append(response_message['content']['data']['text/plain'])
                    elif 'text/html' in response_message['content']['data']:
                        response_type = 'html'
                        response.append(response_message['content']['data']['text/html'])
                    continue

                elif response_message_type == 'status':
                    if response_message['content']['execution_state'] == 'idle':
                        break

        except BaseException as b:
            print(b)

        #finally:
        #    if kernel_socket_future:
        #        try:
        #            kernel_socket_future.close()
        #            kernel_socket_future = None
        #        finally:
        #            # ignore
        #            print('Error closing kernel socket.')

        return '\n'.join(response)


# launcher = KernelLauncher('lresende-elyra:8888')
# kernel = launcher.launch('spark_python_yarn_cluster')
# try:
#     #response = kernel.execute('1 + 1')
#     #pprint (response)
#     response = kernel.execute('print("Hello World")')
#     pprint (response)
# finally:
#     launcher.shutdown(kernel.kernel_id)
#




# -*- coding: utf-8 -*-

"""Jupyter Enterprise Scheduler - Schedule Notebook execution."""
import os
import sys

import asyncio
import click
from flask import Flask
from flask_restful import Api

from enterprise_scheduler.scheduler_resource import SchedulerResource
from enterprise_scheduler.util import fix_asyncio_event_loop_policy

server_name = os.getenv('SERVER_NAME','127.0.0.1:5000')

@click.command()
@click.option('--gateway_host', default='lresende-elyra:8888', help='Jupyter Enterprise Gateway host information')
@click.option('--kernelspec', default='python2', help='Jupyter Notebook kernelspec to use while executing notebook')
def main(gateway_host, kernelspec):
    """Jupyter Enterprise Scheduler - Schedule Notebook execution."""
    click.echo('Starting Scheduler at {} using Gateway at {} with default kernelspec {}'.format(server_name, gateway_host, kernelspec))
    click.echo('Add new tasks via post commands to http://{}/scheduler/tasks '.format(server_name))

    fix_asyncio_event_loop_policy(asyncio)

    app = Flask('Notebook Scheduler')
    api = Api(app)

    api.add_resource(SchedulerResource, '/scheduler/tasks',
                     resource_class_kwargs={ 'default_gateway_host': gateway_host, 'default_kernelspec': kernelspec })

    print('Add new tasks via http://{}/scheduler/tasks '.format(server_name))

    server_parts = server_name.split(':')

    app.run(host=server_parts[0], port=int(server_parts[1]), debug=True, use_reloader=False)

    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import queue
import uuid
from threading import Thread
from urllib.request import urlopen

from enterprise_scheduler.executor import JupyterExecutor, FfDLExecutor


class Scheduler:

    def __init__(self, default_gateway_host=None, default_kernelspec=None, number_of_threads=5):
        self.default_gateway_host = default_gateway_host
        self.default_kernelspec = default_kernelspec
        self.number_of_threads = number_of_threads

        self.executors = {}
        self.executors[JupyterExecutor.TYPE] = JupyterExecutor()
        self.executors[FfDLExecutor.TYPE] = FfDLExecutor()

        self.queue = queue.PriorityQueue()
        self.executor_threads = []
        self.running = False

    def _executor(self):
        while self.running:
            if not self.queue.empty():
                task = self.queue.get()
                self._execute_task(task)
                self.queue.task_done()

    def _execute_task(self, task):
        executor_type = task['executor'].lower()  # Jupyter, Docker, FfDL
        executor = self.executors[executor_type]
        executor.execute_task(task)

    def schedule_task(self, task):
        id = uuid.uuid4()
        task['id'] = id

        if 'notebook_location' in task.keys():
            notebook_location = task['notebook_location']
            task['notebook'] = self._read_remote_notebook_content(notebook_location)

        self._validate_task(task)

        print('adding task [{}] to queue:\n {}'.format(id, str(task)))
        self.queue.put(item=task)
        return id

    def start(self):
        self.running = True
        for i in range(1, self.number_of_threads):
            t = Thread(target=self._executor)
            t.daemon = True

            self.executor_threads.append(t)

            t.start()

    def stop(self):
        #self.queue.join()
        self.running = False

        for t in self.executor_threads:
            t.join()

    @staticmethod
    def _validate_task(task):
        if 'executor' not in task.keys():
            raise ValueError('Submitted task is missing [executor] information')

        if 'endpoint' not in task.keys():
            raise ValueError('Submitted task is missing [endpoint] information')

        if 'kernelspec' not in task.keys():
            raise ValueError('Submitted task is missing [kernelspec] information')

        if 'notebook_location' not in task.keys() and 'notebook' not in task.keys():
            raise ValueError('Submitted task is missing notebook information (either notebook_location or notebook)')


    @staticmethod
    def _read_remote_notebook_content(notebook_location):
        try:
            notebook_content = urlopen(notebook_location).read().decode()
            return notebook_content
        except BaseException as base:
            raise Exception('Error reading notebook source "{}": {}'.format(notebook_location, base))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `enterprise_scheduler` package."""

import asyncio
import os
import json
import unittest
import time
from pprint import pprint

from enterprise_scheduler.scheduler import Scheduler
from enterprise_scheduler.util import fix_asyncio_event_loop_policy

RESOURCES = os.path.join(os.path.dirname(__file__), 'resources')

DEFAULT_GATEWAY = "lresende-elyra:8888"
DEFAULT_KERNELSPEC = "python2"


class TestEnterpriseScheduler(unittest.TestCase):
    """Tests for `enterprise_scheduler` package."""
    scheduler = None

    @classmethod
    def setUpClass(cls):
        """Set up test fixtures, if any."""
        fix_asyncio_event_loop_policy(asyncio)

        cls.scheduler = Scheduler()
        cls.scheduler.start()

    @classmethod
    def tearDownClass(cls):
        """Tear down test fixtures, if any."""
        cls.scheduler.stop()
        cls.scheduler = None

    def test_execute_jupyter_task_with_embedded_notebook(self):
        notebook = self._read_notebook('simple.ipynb')

        task = {}
        task['executor'] = 'jupyter'
        task['endpoint'] = DEFAULT_GATEWAY
        task['kernelspec'] = DEFAULT_KERNELSPEC
        task['notebook'] = notebook

        TestEnterpriseScheduler.scheduler.schedule_task(task)

    def test_execute_jupyter_task_with_remote_notebook(self):
        task = {}
        task['executor'] = 'jupyter'
        task['endpoint'] = DEFAULT_GATEWAY
        task['kernelspec'] = DEFAULT_KERNELSPEC
        task['notebook_location'] = 'http://home.apache.org/~lresende/notebooks/notebook-brunel.ipynb'

        TestEnterpriseScheduler.scheduler.schedule_task(task)

    def test_execute_ffdl_task_with_embedded_notebook(self):
        notebook = self._read_notebook('ffdl.ipynb')

        task = {}
        task['executor'] = 'ffdl'
        task['framework'] = 'tensorflow'
        task['endpoint'] = '##########'
        task['kernelspec'] = 'python3'
        task['user'] = '##########'
        task['userinfo'] = 'bluemix-instance-id=test-user'
        task['cpus'] = 1
        task['gpus'] = 0
        task['memory'] = '1Gb'
        task['cos_endpoint'] = '##########'
        task['cos_user'] = '##########'
        task['cos_password'] = '##########'
        task['notebook'] = notebook

        TestEnterpriseScheduler.scheduler.schedule_task(task)


    def _read_notebook(self, filename):
        filename = os.path.join(RESOURCES, filename)
        with open(filename, 'r') as f:
            return json.load(f)



import os
import json
import tempfile
import time
import yaml

import nbformat

from subprocess import Popen, PIPE
from shutil import copyfile

from enterprise_scheduler.kernel_client import KernelLauncher
from enterprise_scheduler.util import zip_directory


class Executor:
    """Base executor class for :
        - Jupyter
        - FFDL (Fabric for Deep Learning)
        - DLAAS (Deep Learning as a Service)"""
    def __init__(self, default_gateway_host=None, default_kernelspec=None):
        self.default_gateway_host = default_gateway_host
        self.default_kernelspec = default_kernelspec


class JupyterExecutor(Executor):
    TYPE = "jupyter"

    def execute_task(self, task):
        # start notebook
        print('')
        print('Start notebook execution...')
        print('Starting kernel...')

        launcher = KernelLauncher(task['endpoint'])
        kernel = launcher.launch(task['kernelspec'])

        time.sleep(10)

        # execute all cells
        try:
            print('reading notebook contents')
            notebook = nbformat.reads(json.dumps(task['notebook']), as_version=4)

            print('Starting cell execution')
            for cell in notebook.cells:
                print('Executing cell\n{}'.format(cell.source))
                response = kernel.execute(cell.source)
                print('Response\n{}'.format(response))
                outputs = []
                outputs.append(response)
                cell['outputs'] = outputs

        except BaseException as base:
            print('Error executing notebook cells: {}'.format(base))

        finally:
            print('Starting kernel shutdown')
            # shutdown notebook
            launcher.shutdown(kernel.kernel_id)

        print('Notebook execution done')
        print('')


class FfDLExecutor(Executor):
    """FFDL Executor Supports :
        - TensorFlow
        - Keras
        - Caffe
        - Caffe 2
        - PyTorch"""
    TYPE = "ffdl"

    def __init__(self):
        self.workdir = os.path.join(tempfile.gettempdir(), str(FfDLExecutor.TYPE))
        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)

        rootdir = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
        self.runtimedir = os.path.join(rootdir, 'resources/ffdl')
        print(self.runtimedir)

    def execute_task(self, task):

        ffdl_endpoint = task['endpoint']
        ffdl_authorization = task['user']
        ffdl_userinfo = task['userinfo']
        ffdl_zip = self._create_ffdl_zip(task)
        ffdl_manifest = self._create_manifest(task)

        command = 'curl -X POST "' + ffdl_endpoint + '" ' \
                  '-H "accept: application/json" ' \
                  '-H "Authorization: ' + ffdl_authorization + '" ' \
                  '-H "X-Watson-Userinfo: ' + ffdl_userinfo + '" ' \
                  '-H "Content-Type: multipart/form-data" ' \
                  '-F "model_definition=@' + ffdl_zip + ';type=application/zip" ' \
                  '-F "manifest=@' + ffdl_manifest + '"'

        #print('>>> Submitting \n {}'.format(command))

        try:
            p = Popen([command ], stdout=PIPE, stderr=PIPE, shell=True, cwd=self.runtimedir)
            p.wait()
            #print(p.stdout.read())
            #print(p.stderr.read())
        finally:
            if(p.stderr):
                p.stderr.close()
            if(p.stdout):
                p.stdout.close()

    def _create_manifest(self, task):
        file_name = 'manifest-' + str(task['id'])[:8]
        file_location = self.workdir + '/' + file_name + ".yml"

        manifest_dict = dict(
            name=file_name,
            description='Train Jupyter Notebook model in FfDL',
            version="1.0",
            gpus=task['gpus'],
            cpus=task['cpus'],
            memory=task['memory'],
            learners=1,
            data_stores= [dict(
                id='sl-internal-os',
                type='mount_cos',
                training_data= dict(
                    container='tf_training_data'
                ),
                training_results= dict(
                    container='tf_trained_model'
                ),
                connection= dict(
                    auth_url=task['cos_endpoint'],
                    user_name=task['cos_user'],
                    password=task['cos_password']
                )
            )],
            framework= dict(
                name=task['framework'],
                version='1.5.0-py3',
                command='./start.sh'    ## Run the start script for EG and kernel
            #),
            #evaluation_metrics=dict(
            #    type=''
            )
        )

        with open(file_location, 'w') as outfile:
            yaml.dump(manifest_dict, outfile, default_flow_style=False)

        return file_location

    def _create_ffdl_zip(self, task):
        unique_id = 'ffdl-' + str(task['id'])[:8]
        task_directory = os.path.join(self.workdir, unique_id)
        os.makedirs(task_directory)

        self._write_file(task_directory, "notebook.ipynb", json.dumps(task['notebook']))

        copyfile(os.path.join(self.runtimedir, "start.sh"),
                 os.path.join(task_directory, "start.sh"))

        copyfile(os.path.join(self.runtimedir, "run_notebook.py"),
                 os.path.join(task_directory, "run_notebook.py"))

        zip_file = os.path.join(self.workdir, '{}.zip'.format(unique_id))
        #print('>>> {}'.format(zip_file))
        #print('>>> {}'.format(task_directory))
        zip_directory(zip_file, task_directory)

        return zip_file


    @staticmethod
    def _write_file(directory, filename, contents):
        filename = os.path.join(directory, filename)
        with open(filename, 'w') as f:
            f.write(str(contents))



import os
import asyncio
import zipfile


def fix_asyncio_event_loop_policy(asyncio):
    """
    Work around https://github.com/tornadoweb/tornado/issues/2183
    """

    class PatchedDefaultEventLoopPolicy(asyncio.DefaultEventLoopPolicy):

        def get_event_loop(self):
            """Get the event loop.
            This may be None or an instance of EventLoop.
            """
            try:
                return super().get_event_loop()
            except RuntimeError:
                # "There is no current event loop in thread"
                loop = self.new_event_loop()
                self.set_event_loop(loop)
                return loop

    asyncio.set_event_loop_policy(PatchedDefaultEventLoopPolicy())

def zip_directory(zip_name, directory):
    zip_file = zipfile.ZipFile(zip_name, 'w', zipfile.ZIP_DEFLATED)

    for root, dirs, files in os.walk(directory):
        for file in files:
            zip_file.write(os.path.join(root, file), file)

    zip_file.close()





from flask import Response, request
from flask_restful import Resource

from enterprise_scheduler.scheduler import Scheduler

scheduler = Scheduler()
scheduler.start()

class SchedulerResource(Resource):
    """
    Scheduler REST API used to submit Jupyter Notebooks for batch executions

    curl -X POST -v http://localhost:5000/scheduler/tasks -d "{\"notebook_location\":\"http://home.apache.org/~lresende/notebooks/notebook-brunel.ipynb\"}"
    """

    def __init__(self, default_gateway_host, default_kernelspec):
        self.default_gateway_host = default_gateway_host
        self.default_kernelspec = default_kernelspec

    def _html_response(self, data):
        resp = Response(data, mimetype='text/plain', headers=None)
        resp.status_code = 200
        return resp

    def post(self):
        global scheduler
        task = request.get_json(force=True)

        if 'endpoint' not in task.keys():
            task['endpoint'] = self.default_gateway_host

        if 'kernelspec' not in task.keys():
            task['kernelspec'] = self.default_kernelspec

        scheduler.schedule_task(task)

        return 'submitted', 201





#!/usr/bin/env python

import time
import nbformat

from enterprise_scheduler.kernel_client import KernelLauncher

print('')
print('Waiting Enterprise Gateway to load...')

time.sleep(10)

print('')
print('Start notebook execution...')

try:
    # start notebook
    print('reading notebook contents')
    notebook = None
    with open('notebook.ipynb', 'r') as f:
        notebook = nbformat.reads(f.read(), as_version=4)

    if not notebook:
        raise RuntimeError('Error reading notebook file')

    print('Starting kernel...')

    launcher = KernelLauncher('localhost:8888')
    kernel = launcher.launch(notebook['metadata']['kernelspec']['name'])

    if not kernel:
        raise RuntimeError('Error starting kernel with kernelspec: {}'.format())

    time.sleep(10)

    print('Starting cell execution')
    for cell in notebook.cells:
        print('Executing cell\n{}'.format(cell.source))
        response = kernel.execute(cell.source)
        print('Response\n{}'.format(response))
        outputs = []
        outputs.append(response)
        cell['outputs'] = outputs


except BaseException as base:
    print('Error executing notebook cells: {}'.format(base))

finally:
    print('Starting kernel shutdown')
    # shutdown notebook
    launcher.shutdown(kernel.kernel_id)

print('Notebook execution done')
print('')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    'click>=6.0',
    'bumpversion>=0.5.3',
    'wheel>=0.30.0',
    'watchdog>=0.8.3',
    'flake8>=3.5.0',
    'tox>=2.9.1',
    'coverage>=4.5.1',
    'Sphinx>=1.7.1',
    'twine>=1.10.0',
    'nbconvert>=5.3.1',
    'requests >= 2.8, < 3.0',
    'flask-restful>=0.3.6'
]

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="Luciano Resende",
    author_email='lresende@apache.org',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    description="Python Boilerplate contains all the boilerplate you need to create a Python package.",
    entry_points={
        'console_scripts': [
            'enterprise_scheduler=enterprise_scheduler.scheduler_application:main',
        ],
    },
    install_requires=requirements,
    license="Apache license",
    long_description=readme,
    include_package_data=True,
    keywords='enterprise_scheduler',
    name='enterprise_scheduler',
    packages=find_packages(include=['enterprise_scheduler']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/lresende/enterprise_scheduler',
    version='0.1.0.dev2',
    zip_safe=False,
)
